@@ -2,7 +2,9 @@
 // SPDX-License-Identifier: Apache-2.0
 
 #include <openvdb/openvdb.h>
+#include <openvdb/io/File.h>
 #include <openvdb/points/PointRasterizeSDF.h>
+#include <openvdb/points/PrincipalComponentAnalysis.h>
 #include "PointBuilder.h"
 
 #include <gtest/gtest.h>
@@ -18,9 +20,14 @@ public:
 
 
 template <typename FilterT>
-struct FixedSpheres
+struct FixedSurfacing
 {
-    FixedSpheres(const FilterT& f = FilterT()) : filter(f) {}
+    /// @note  The surface and surfaceSmooth methods use the old API. Once this
+    ///   has been deprecated they can be swicthed over to the new AI (as
+    ///   demonstrated by surfaceEllips).
+    FixedSurfacing(const FilterT& f = FilterT()) : filter(f) {}
+
+    //
 
     FloatGrid::Ptr surface(const Real radius)
     {
@@ -38,6 +45,8 @@ struct FixedSpheres
         return grids;
     }
 
+    //
+
     FloatGrid::Ptr surfaceSmooth(const Real radius, const Real search)
     {
         auto sdf = points::rasterizeSmoothSpheres(*points, radius, search, halfband, transform, filter);
@@ -61,9 +70,9 @@ struct FixedSpheres
 };
 
 template <typename FilterT>
-struct VariableSpheres : public FixedSpheres<FilterT>
+struct VariableSurfacing : public FixedSurfacing<FilterT>
 {
-    VariableSpheres(const FilterT& f = FilterT()) : FixedSpheres<FilterT>(f) {}
+    VariableSurfacing(const FilterT& f = FilterT()) : FixedSurfacing<FilterT>(f) {}
 
     FloatGrid::Ptr surface(const Real scale = 1.0, const std::string& pscale = "pscale")
     {
@@ -106,7 +115,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSpheres)
     // Test no points
     {
         float radius = 0.2f;
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({}).voxelsize(0.1).get();
         s.transform = nullptr;
@@ -121,7 +130,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSpheres)
 
     // Test single point
     {
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
 
         // small radius, small voxel size
         float radius = 0.2f;
@@ -221,7 +230,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSpheres)
 
     // Test multiple points - 8 points at cube corner positions
     {
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
         float radius = 0.2f;
         auto positions = getBoxPoints(/*scale*/0.0f);
 
@@ -348,7 +357,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSpheres)
             .group({1,0,1,0,1,0,1,0}, "test")
             .get();
         points::GroupFilter filter("test", points->tree().cbeginLeaf()->attributeSet());
-        FixedSpheres<points::GroupFilter> s(filter);
+        FixedSurfacing<points::GroupFilter> s(filter);
         s.halfband = 3;
         s.points = points;
         s.transform = nullptr;
@@ -380,13 +389,13 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSpheres)
 }
 
 
-TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
+TEST_F(TestPointRasterizeSDF, testRasterizeVariableSurfacing)
 {
     // First few tests check that the results are fp equivalent to fixed spheres
 
     // Test no points
     {
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({}).voxelsize(0.1).get();
         s.transform = nullptr;
@@ -401,7 +410,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
 
     // Test single point
     {
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         float radius = 0.2f;
 
         // small radius, small voxel size
@@ -410,7 +419,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
         s.transform = nullptr;
 
         FloatGrid::Ptr sdf = s.surface();
-        FloatGrid::Ptr comp = s.FixedSpheres<points::NullFilter>::surface(radius);
+        FloatGrid::Ptr comp = s.FixedSurfacing<points::NullFilter>::surface(radius);
         EXPECT_TRUE(sdf && sdf->getName() == "variable");
         EXPECT_TRUE(comp && comp->getName() == "fixed");
         EXPECT_TRUE(sdf->tree().hasSameTopology(comp->tree()));
@@ -426,7 +435,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
         s.transform = nullptr;
 
         sdf = s.surface();
-        comp = s.FixedSpheres<points::NullFilter>::surface(radius);
+        comp = s.FixedSurfacing<points::NullFilter>::surface(radius);
         EXPECT_TRUE(sdf && sdf->getName() == "variable");
         EXPECT_TRUE(comp && comp->getName() == "fixed");
         EXPECT_TRUE(sdf->tree().hasSameTopology(comp->tree()));
@@ -443,7 +452,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
         s.transform = math::Transform::createLinearTransform(0.3);
 
         sdf = s.surface();
-        comp = s.FixedSpheres<points::NullFilter>::surface(radius);
+        comp = s.FixedSurfacing<points::NullFilter>::surface(radius);
         EXPECT_TRUE(sdf && sdf->getName() == "variable");
         EXPECT_TRUE(comp && comp->getName() == "fixed");
         EXPECT_TRUE(sdf->tree().hasSameTopology(comp->tree()));
@@ -455,7 +464,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
 
     // Test multiple points - 8 points at cube corner positions
     {
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         float radius = 0.2f;
         auto positions = getBoxPoints(/*scale*/0.0f);
         // test points overlapping all at 0,0,0 - should produce the same grid as first test
@@ -464,7 +473,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
         s.transform = nullptr;
 
         FloatGrid::Ptr sdf = s.surface();
-        FloatGrid::Ptr comp = s.FixedSpheres<points::NullFilter>::surface(radius);
+        FloatGrid::Ptr comp = s.FixedSurfacing<points::NullFilter>::surface(radius);
         EXPECT_TRUE(sdf && sdf->getName() == "variable");
         EXPECT_TRUE(comp && comp->getName() == "fixed");
         EXPECT_TRUE(sdf->tree().hasSameTopology(comp->tree()));
@@ -481,7 +490,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
         s.transform = nullptr;
 
         sdf = s.surface();
-        comp = s.FixedSpheres<points::NullFilter>::surface(radius);
+        comp = s.FixedSurfacing<points::NullFilter>::surface(radius);
         EXPECT_TRUE(sdf && sdf->getName() == "variable");
         EXPECT_TRUE(comp && comp->getName() == "fixed");
         EXPECT_TRUE(sdf->tree().hasSameTopology(comp->tree()));
@@ -508,7 +517,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
         s.transform = math::Transform::createLinearTransform(mat);
 
         sdf = s.surface();
-        comp = s.FixedSpheres<points::NullFilter>::surface(radius);
+        comp = s.FixedSurfacing<points::NullFilter>::surface(radius);
         EXPECT_TRUE(sdf && sdf->getName() == "variable");
         EXPECT_TRUE(comp && comp->getName() == "fixed");
         EXPECT_TRUE(sdf->tree().hasSameTopology(comp->tree()));
@@ -530,13 +539,13 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
             .attribute(radius, "pscale")
             .get();
         points::GroupFilter filter("test", points->tree().cbeginLeaf()->attributeSet());
-        VariableSpheres<points::GroupFilter> s(filter);
+        VariableSurfacing<points::GroupFilter> s(filter);
         s.halfband = 3;
         s.points = points;
         s.transform = nullptr;
 
         FloatGrid::Ptr sdf = s.surface();
-        FloatGrid::Ptr comp = s.FixedSpheres<points::GroupFilter>::surface(radius);
+        FloatGrid::Ptr comp = s.FixedSurfacing<points::GroupFilter>::surface(radius);
         EXPECT_TRUE(sdf && sdf->getName() == "variable");
         EXPECT_TRUE(comp && comp->getName() == "fixed");
         EXPECT_TRUE(sdf->tree().hasSameTopology(comp->tree()));
@@ -557,7 +566,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSpheres)
             .attribute(rads, "myrad")
             .get();
 
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         s.halfband = 1; // small half band
         s.points = points;
         s.transform = nullptr;
@@ -609,7 +618,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSmoothSpheres)
     // Test no points
     {
         float radius = 0.2f, search = 0.4f;
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({}).voxelsize(0.1).get();
         s.transform = nullptr;
@@ -624,7 +633,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSmoothSpheres)
 
     // Test single point
     {
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({ Vec3f(0) }).voxelsize(0.1).get();
         s.transform = nullptr;
@@ -676,7 +685,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSmoothSpheres)
             .group({1,0,0,0,0,0,0,0}, "test") // only first point
             .get();
         points::GroupFilter filter("test", points->tree().cbeginLeaf()->attributeSet());
-        FixedSpheres<points::GroupFilter> s(filter);
+        FixedSurfacing<points::GroupFilter> s(filter);
         s.halfband = 3;
         s.points = points;
         s.transform = nullptr;
@@ -697,9 +706,51 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSmoothSpheres)
         }
     }
 
+    // Test two points qhich create a ghost particle outside of their
+    // radii (to test that the surface topology correctly accounts for the
+    // search distance)
+    {
+        FixedSurfacing<points::NullFilter> s;
+        s.halfband = 2;
+        s.points = PointBuilder({ Vec3f(0), Vec3f(0,5,0) }).voxelsize(0.1).get();
+        s.transform = nullptr;
+        double radius = 0.5, search = 5; // large search
+
+        FloatGrid::Ptr sdf = s.surfaceSmooth(radius, search);
+        EXPECT_TRUE(sdf && sdf->getName() == "fixed_avg");
+        EXPECT_TRUE(sdf->transform() == s.points->transform());
+        EXPECT_EQ(GRID_LEVEL_SET, sdf->getGridClass());
+        EXPECT_EQ(float(s.halfband * s.points->voxelSize()[0]), sdf->background());
+
+        // @todo  regression test. find a way to better test these values
+        size_t interiorOn = 0, exteriorOn = 0;
+        EXPECT_EQ(Index32(36), sdf->tree().leafCount());
+        EXPECT_EQ(Index64(0), sdf->tree().activeTileCount());
+        EXPECT_EQ(Index64(4188), sdf->tree().activeVoxelCount());
+
+        for (auto iter = sdf->cbeginValueOn(); iter; ++iter) {
+            EXPECT_TRUE(*iter > -sdf->background());
+            EXPECT_TRUE(*iter < sdf->background());
+            if (*iter > 0) ++exteriorOn;
+            else           ++interiorOn;
+        }
+        EXPECT_EQ(size_t(1244), interiorOn);
+        EXPECT_EQ(size_t(2944), exteriorOn);
+
+        size_t interiorOff = 0, exteriorOff = 0;
+        for (auto iter = sdf->cbeginValueOff(); iter; ++iter) {
+            EXPECT_TRUE((*iter <= -sdf->background()) || (*iter >= sdf->background()))
+                << *iter << " " << sdf->background();;
+            if (*iter > 0) ++exteriorOff;
+            else           ++interiorOff;
+        }
+        EXPECT_EQ(size_t(323), interiorOff);
+        EXPECT_EQ(size_t(308789), exteriorOff);
+    }
+
     // Test multiple points - 8 points at cube corner positions
     {
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
         // test points overlapping all at 0,0,0 - should produce the same grid as first test
         auto positions = getBoxPoints(/*scale*/0.0f);
         s.halfband = 3;
@@ -719,7 +770,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeSmoothSpheres)
 
         // test points from a box with coords at 0.5 and a search radius
         // large enough to create a smoothed box
-        // @todo find a way to better test these values
+        // @todo  regression test. find a way to better test these values
         positions = getBoxPoints(/*scale*/0.5f);
         radius = 0.2, search = 1.2;
         s.halfband = 3;
@@ -756,7 +807,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSmoothSpheres)
     // Test no points
     {
         float radius = 0.2f, search = 0.4f;
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({}).voxelsize(0.1).get();
         s.transform = nullptr;
@@ -771,7 +822,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSmoothSpheres)
 
     // Test single point
     {
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({ Vec3f(0) }).voxelsize(0.1).attribute(0.2f, "rad").get();
         s.transform = math::Transform::createLinearTransform(0.1);
@@ -791,7 +842,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSmoothSpheres)
         search = 0.6;
 
         sdf = s.surfaceSmooth(scale, search, "rad");
-        FloatGrid::Ptr comp = s.FixedSpheres<points::NullFilter>::surfaceSmooth(0.2, search);
+        FloatGrid::Ptr comp = s.FixedSurfacing<points::NullFilter>::surfaceSmooth(0.2, search);
         EXPECT_TRUE(sdf && sdf->getName() == "variable_avg");
         EXPECT_TRUE(comp && comp->getName() == "fixed_avg");
         EXPECT_TRUE(sdf->transform() == *s.transform);
@@ -810,7 +861,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSmoothSpheres)
         scale = 0.5; // 0.4*0.5 = 0.2
         search = 5; // search of 5 allows for halfband size to up 10
         sdf = s.surfaceSmooth(scale, search, "rad");
-        comp = s.FixedSpheres<points::NullFilter>::surfaceSmooth(0.2, search);
+        comp = s.FixedSurfacing<points::NullFilter>::surfaceSmooth(0.2, search);
         EXPECT_TRUE(sdf && sdf->getName() == "variable_avg");
         EXPECT_TRUE(comp && comp->getName() == "fixed_avg");
         EXPECT_TRUE(sdf->transform() == *s.transform);
@@ -833,7 +884,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSmoothSpheres)
             .attribute(float(radius), "pscale")
             .get();
         points::GroupFilter filter("test", points->tree().cbeginLeaf()->attributeSet());
-        VariableSpheres<points::GroupFilter> s(filter);
+        VariableSurfacing<points::GroupFilter> s(filter);
         s.halfband = 3;
         s.points = points;
         s.transform = nullptr;
@@ -864,7 +915,7 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSmoothSpheres)
         std::vector<float> rads = {1.1f, 1.3f, 1.5f, 1.7f, 2.1f, 2.3f, 2.5f, 2.7f};
         double scale = 0.6, search = 2.0;
 
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         s.halfband = 4; // large enough to fill interior of the cube
         s.points = PointBuilder(positions).voxelsize(0.2).attribute(rads, "myrad").get();
         s.transform = nullptr;
@@ -888,13 +939,381 @@ TEST_F(TestPointRasterizeSDF, testRasterizeVariableSmoothSpheres)
     }
 }
 
+TEST_F(TestPointRasterizeSDF, testRasterizeEllipsoids)
+{
+    // @brief  Small helper struct to represent a single ellipsoid
+    struct Ellipse
+    {
+        Ellipse(const Vec3d center, const Vec3d stretch, const Mat3s rotation)
+            : mCenter(center), mStretch(stretch), mRotation(rotation)
+            , mInverseRotation(mRotation.transpose())
+            , mInverseTransform([&](){
+                // construct inverse transformation to create sphere out of an ellipsoid
+                // The transform that defines how we go from each voxel back to the source point
+                const Vec3d inverseStretch = 1.0f / mStretch;
+                math::Mat3s invDiag;
+                invDiag.setSymmetric(inverseStretch, Vec3f(0));
+                return invDiag * mInverseRotation;
+            }())
+            , mMinRadius(std::min(mStretch.x(), std::min(mStretch.y(), mStretch.z()))) {}
+
+#if OPENVDB_ELLIPSOID_KERNEL_MODE == 1
+        double project(const Vec3d P) const
+        {
+            const Vec3d dir = (P-mCenter);
+            const Vec3d pUnitSphere = mInverseTransform * dir;
+            const double len = math::Sqrt(pUnitSphere.lengthSqr());
+
+            if (len == 0) { // exactly at center
+                return -mMinRadius;
+            }
+            else {
+                Vec3d ellipsNormal = (mInverseTransform.transpose() * pUnitSphere);
+                ellipsNormal.normalize();
+                // project back
+                return static_cast<double>(dir.dot(ellipsNormal) * (1.0 - (float(1.0)/len)));
+            }
+        }
+#elif OPENVDB_ELLIPSOID_KERNEL_MODE == 2
+        double project(const Vec3d P) const
+        {
+            const Vec3d dir = (P-mCenter);
+            const Vec3d pUnitSphere = mInverseRotation * dir;
+            const Vec3d radInv = 1.0f / mStretch;
+            const Vec3d radInv2 = 1.0f / math::Pow2(mStretch);
+            const double k1 = (pUnitSphere * radInv).length();
+            if (k1 == 0) { // exactly at center
+                return -mMinRadius;
+            }
+            const double k2 = (pUnitSphere * radInv2).length();
+            return static_cast<double>((k1 * (k1 - double(1.0)) / k2));
+        }
+#endif
+
+    private:
+        Vec3d mCenter;
+        Vec3d mStretch;
+        math::Mat3s mRotation;
+        math::Mat3s mInverseRotation;
+        math::Mat3s mInverseTransform;
+        double mMinRadius;
+    };
+
+    //
+
+    const points::PcaAttributes pcaAttrs;
+
+    // Test no points
+    {
+        points::EllipsoidSettings<> s;
+        s.radiusScale = Vec3f(0.2f);
+        s.halfband = 3;
+        s.transform = nullptr;
+
+        auto points = PointBuilder({}).voxelsize(0.1).get();
+        auto grids = points::rasterizeSdf(*points, s);
+        FloatGrid::Ptr sdf = StaticPtrCast<FloatGrid>(grids.front());
+
+        EXPECT_TRUE(sdf);
+        EXPECT_TRUE(sdf->transform() == points->transform());
+        EXPECT_EQ(GRID_LEVEL_SET, sdf->getGridClass());
+        EXPECT_EQ(float(s.halfband * points->voxelSize()[0]), sdf->background());
+        EXPECT_TRUE(sdf->empty());
+    }
+
+    // Test single point which is not treated as an ellips. This should give
+    // identicle results to SphereSettings<>
+    {
+        points::EllipsoidSettings<> s;
+        s.radiusScale = Vec3f(0.2f);
+        s.halfband = 3;
+        s.transform = nullptr;
+        s.radius = pcaAttrs.stretch;
+        s.rotation = pcaAttrs.rotation;
+        s.pws = pcaAttrs.positionWS;
+
+        /// 1) test with a single point with uniform stretch
+        auto points = PointBuilder({Vec3f(0)})
+            .voxelsize(0.1)
+            .attribute(points::PcaAttributes::StretchT(1.0), pcaAttrs.stretch) // uniform stretch
+            .attribute(points::PcaAttributes::RotationT::identity(), pcaAttrs.rotation)
+            .attribute(points::PcaAttributes::PosWsT(0.0), pcaAttrs.positionWS)
+            .get();
+
+        auto grids = points::rasterizeSdf(*points, s);
+        FloatGrid::Ptr sdf = StaticPtrCast<FloatGrid>(grids.front());
+        EXPECT_TRUE(sdf);
+        EXPECT_TRUE(sdf->transform() == points->transform());
+        EXPECT_EQ(GRID_LEVEL_SET, sdf->getGridClass());
+        EXPECT_EQ(float(s.halfband * points->voxelSize()[0]), sdf->background());
+
+        EXPECT_EQ(Index32(8), sdf->tree().leafCount());
+        EXPECT_EQ(Index64(0), sdf->tree().activeTileCount());
+        EXPECT_EQ(Index64(485), sdf->tree().activeVoxelCount());
+
+        for (auto iter = sdf->cbeginValueOn(); iter; ++iter) {
+            const Vec3d ws = sdf->transform().indexToWorld(iter.getCoord());
+            float length = float(ws.length()); // dist to center
+            length -= float(s.radiusScale.x()); // account for radius
+            EXPECT_NEAR(length, *iter, 1e-6f);
+        }
+
+        // should only have exterior background
+        for (auto iter = sdf->cbeginValueOff(); iter; ++iter) {
+            EXPECT_EQ(sdf->background(), *iter);
+        }
+
+        // should only have exterior background
+        for (auto iter = sdf->cbeginValueOff(); iter; ++iter) {
+            EXPECT_EQ(sdf->background(), *iter);
+        }
+
+        /// 2) larger radius, larger voxel size
+        s.radiusScale = Vec3f(1.3f);
+        float stretch = 1.6f;
+        points = PointBuilder({Vec3f(0)})
+            .voxelsize(0.5)
+            .attribute(points::PcaAttributes::StretchT(stretch), pcaAttrs.stretch) // uniform stretch
+            .attribute(points::PcaAttributes::RotationT::identity(), pcaAttrs.rotation)
+            .attribute(points::PcaAttributes::PosWsT(0.0), pcaAttrs.positionWS)
+            .get();
+
+        grids = points::rasterizeSdf(*points, s);
+        sdf = StaticPtrCast<FloatGrid>(grids.front());
+
+        EXPECT_TRUE(sdf);
+        EXPECT_TRUE(sdf->transform() == points->transform());
+        EXPECT_EQ(GRID_LEVEL_SET, sdf->getGridClass());
+        EXPECT_EQ(float(s.halfband * points->voxelSize()[0]), sdf->background());
+
+        EXPECT_EQ(Index32(8), sdf->tree().leafCount());
+        EXPECT_EQ(Index64(0), sdf->tree().activeTileCount());
+        EXPECT_EQ(Index64(1544), sdf->tree().activeVoxelCount());
+
+        for (auto iter = sdf->cbeginValueOn(); iter; ++iter) {
+            const Vec3d ws = sdf->transform().indexToWorld(iter.getCoord());
+            float length = float(ws.length()); // dist to center
+            length -= float(s.radiusScale.x() * stretch); // account for radius and sphere scale
+            EXPECT_NEAR(length, *iter, 1e-6f);
+        }
+
+        // check off values
+        size_t interiorOff = 0, exteriorOff = 0;
+        for (auto iter = sdf->cbeginValueOff(); iter; ++iter) {
+            const Vec3d ws = sdf->transform().indexToWorld(iter.getCoord());
+            float length = float(ws.length()); // dist to center
+            // if length is <= the (rad - halfbandws), voxel is inside the surface
+            const bool interior = (length <= ((s.radiusScale.x() * stretch) - (s.halfband * sdf->voxelSize()[0])));
+            if (interior) EXPECT_EQ(-(sdf->background()), *iter);
+            else          EXPECT_EQ(sdf->background(), *iter);
+            interior ? ++interiorOff : ++exteriorOff;
+        }
+
+        EXPECT_EQ(size_t(7), interiorOff);
+        EXPECT_EQ(size_t(297441), exteriorOff);
+
+        /// 3) offset position, different transform, larger half band
+        stretch = 1.0f;
+        s.radiusScale = Vec3f(2.0f);
+        s.halfband = 4;
+        s.transform = math::Transform::createLinearTransform(0.3);
+        points::PcaAttributes::RotationT rot;
+        rot.setToRotation({0,1,0}, 45); // arbitrary rotation should have no effect
+
+        const Vec3f center(-1.2f, 3.4f,-5.6f);
+        points = PointBuilder({Vec3f(center)})
+            .voxelsize(0.1)
+            .attribute(points::PcaAttributes::StretchT(stretch), pcaAttrs.stretch)
+            .attribute(rot, pcaAttrs.rotation)
+            .attribute(points::PcaAttributes::PosWsT(center), pcaAttrs.positionWS)
+            .get();
+
+        grids = points::rasterizeSdf(*points, s);
+        sdf = StaticPtrCast<FloatGrid>(grids.front());
+
+        EXPECT_TRUE(sdf);
+        EXPECT_TRUE(sdf->transform() == *s.transform);
+        EXPECT_EQ(GRID_LEVEL_SET, sdf->getGridClass());
+        EXPECT_EQ(float(s.halfband * s.transform->voxelSize()[0]), sdf->background());
+
+        EXPECT_EQ(Index32(27), sdf->tree().leafCount());
+        EXPECT_EQ(Index64(0), sdf->tree().activeTileCount());
+        EXPECT_EQ(Index64(5005), sdf->tree().activeVoxelCount());
+
+        for (auto iter = sdf->cbeginValueOn(); iter; ++iter) {
+            const Vec3d ws = sdf->transform().indexToWorld(iter.getCoord()) - center;
+            float length = float(ws.length()); // dist to center
+            length -= float(s.radiusScale.x()); // account for radius
+            EXPECT_NEAR(length, *iter, 1e-6f);
+        }
+
+        // check off values
+        interiorOff = 0;
+        exteriorOff = 0;
+        for (auto iter = sdf->cbeginValueOff(); iter; ++iter) {
+            const Vec3d ws = sdf->transform().indexToWorld(iter.getCoord()) - center;
+            float length = float(ws.length()); // dist to center
+            // if length is <= the (rad - halfbandws), voxel is inside the surface
+            const bool interior = (length <= (s.radiusScale.x() - (s.halfband * s.transform->voxelSize()[0])));
+            if (interior) EXPECT_EQ(-sdf->background(), *iter);
+            else          EXPECT_EQ(sdf->background(), *iter);
+            interior ? ++interiorOff : ++exteriorOff;
+        }
+
+        EXPECT_EQ(size_t(80), interiorOff);
+        EXPECT_EQ(size_t(82438), exteriorOff);
+    }
+
+    // Test single point which is treated as an ellips with scale and rotation
+    // along one principal axis
+    {
+        points::EllipsoidSettings<> s;
+        s.radiusScale = Vec3f(0.8f);
+        s.halfband = 3;
+        s.transform = nullptr;
+        s.radius = pcaAttrs.stretch;
+        s.rotation = pcaAttrs.rotation;
+        s.pws = pcaAttrs.positionWS;
+
+        // Design an ellips that is squashed in XYZ and then rotated
+        const points::PcaAttributes::StretchT stretch(1.0f, 0.2f, 1.0f);
+        points::PcaAttributes::RotationT rot;
+         // 45 degree rotation about Y (we only squash in Y so this should be a no-op)
+        rot.setToRotation({0,1,0}, 45);
+
+        /// 1) test with a single ellips with Y stretch/rotation
+        auto points = PointBuilder({Vec3f(0)})
+            .voxelsize(0.1)
+            .attribute(stretch, pcaAttrs.stretch)
+            .attribute(rot, pcaAttrs.rotation)
+            .attribute(points::PcaAttributes::PosWsT(0.0), pcaAttrs.positionWS)
+            .get();
+
+        auto grids = points::rasterizeSdf(*points, s);
+        FloatGrid::Ptr sdf = StaticPtrCast<FloatGrid>(grids.front());
+        EXPECT_TRUE(sdf);
+        EXPECT_TRUE(sdf->transform() == points->transform());
+        EXPECT_EQ(GRID_LEVEL_SET, sdf->getGridClass());
+        EXPECT_EQ(float(s.halfband * points->voxelSize()[0]), sdf->background());
+
+        EXPECT_EQ(Index32(32), sdf->tree().leafCount());
+        EXPECT_EQ(Index64(0), sdf->tree().activeTileCount());
+        EXPECT_EQ(Index64(3705), sdf->tree().activeVoxelCount());
+
+        Ellipse ellipse(Vec3d(0), (stretch * (s.radiusScale/sdf->voxelSize()[0])), rot);
+
+        for (auto iter = sdf->cbeginValueOn(); iter; ++iter) {
+            const float distance = float(ellipse.project(iter.getCoord().asVec3d()) * float(sdf->voxelSize()[0]));
+            EXPECT_NEAR(distance, *iter, 1e-6f);
+        }
+
+        // check off values (because we're also squashing the halfband we
+        // just compre the overal length to 0.0)
+        size_t interiorOff = 0, exteriorOff = 0;
+        for (auto iter = sdf->cbeginValueOff(); iter; ++iter) {
+            const float distance = float(ellipse.project(iter.getCoord().asVec3d()) * float(sdf->voxelSize()[0]));
+            const bool interior = (distance <= 0.0);
+            if (interior) EXPECT_EQ(-sdf->background(), *iter);
+            else          EXPECT_EQ(sdf->background(), *iter);
+            interior ? ++interiorOff : ++exteriorOff;
+        }
+
+        EXPECT_EQ(size_t(80), interiorOff);
+        EXPECT_EQ(size_t(307471), exteriorOff);
+
+        // Run again with a different Y rotation, should be the same as the above
+        // as we're only squashing in Y
+        rot.setToRotation({0,1,0}, -88);
+        points = PointBuilder({Vec3f(0)})
+            .voxelsize(0.1)
+            .group({1}, pcaAttrs.ellipses) // surface as an ellips
+            .attribute(stretch, pcaAttrs.stretch)
+            .attribute(rot, pcaAttrs.rotation)
+            .attribute(points::PcaAttributes::PosWsT(0.0), pcaAttrs.positionWS)
+            .get();
+
+        grids = points::rasterizeSdf(*points, s);
+        FloatGrid::Ptr sdf2 = StaticPtrCast<FloatGrid>(grids.front());
+        EXPECT_TRUE(sdf2);
+        EXPECT_TRUE(sdf2->transform() == points->transform());
+        EXPECT_EQ(GRID_LEVEL_SET, sdf2->getGridClass());
+        EXPECT_EQ(float(s.halfband * points->voxelSize()[0]), sdf2->background());
+        EXPECT_TRUE(sdf->tree().hasSameTopology(sdf2->tree()));
+
+        for (auto iter = sdf->cbeginValueOn(); iter; ++iter) {
+            EXPECT_NEAR(sdf2->tree().getValue(iter.getCoord()), *iter, 1e-6f);
+        }
+    }
+
+    // Test single point which is treated as an ellips with scale and rotation
+    // along multiple axis
+    {
+        points::EllipsoidSettings<> s;
+        s.halfband = 5;
+        s.radius = pcaAttrs.stretch;
+        s.rotation = pcaAttrs.rotation;
+        s.pws = pcaAttrs.positionWS;
+
+        // Design an ellips that is squashed in XYZ and then rotated
+        const points::PcaAttributes::StretchT stretch(0.3f, 0.6f, 1.8f);
+        points::PcaAttributes::RotationT a,b,c,rot;
+        a.setToRotation({1,0,0}, 20);
+        b.setToRotation({0,1,0}, 45);
+        c.setToRotation({0,0,1}, 66);
+        rot = a * b * c;
+
+        //
+
+        /// 1) test with a single ellips with Y stretch/rotation
+        const Vec3f center(-1.2f, 3.4f,-5.6f);
+        auto points = PointBuilder({center})
+            .voxelsize(0.2)
+            .attribute(stretch, pcaAttrs.stretch)
+            .attribute(rot, pcaAttrs.rotation)
+            .attribute(points::PcaAttributes::PosWsT(center), pcaAttrs.positionWS)
+            .get();
+
+        auto grids = points::rasterizeSdf(*points, s);
+        FloatGrid::Ptr sdf = StaticPtrCast<FloatGrid>(grids.front());
+        EXPECT_TRUE(sdf);
+        EXPECT_TRUE(sdf->transform() == points->transform());
+        EXPECT_EQ(GRID_LEVEL_SET, sdf->getGridClass());
+        EXPECT_EQ(float(s.halfband * points->voxelSize()[0]), sdf->background());
+
+        EXPECT_EQ(Index32(32), sdf->tree().leafCount());
+        EXPECT_EQ(Index64(0), sdf->tree().activeTileCount());
+        EXPECT_EQ(Index64(4715), sdf->tree().activeVoxelCount());
+
+        const Ellipse ellipse(sdf->worldToIndex(center), stretch * (1/sdf->voxelSize()[0]), rot);
+
+        for (auto iter = sdf->cbeginValueOn(); iter; ++iter) {
+            const float distance = float(ellipse.project(iter.getCoord().asVec3d()) * float(sdf->voxelSize()[0]));
+            EXPECT_NEAR(distance, *iter, 1e-6f);
+        }
+
+        // check off values (because we're also squashing the halfband we
+        // just compre the overal length to 0.0)
+        size_t interiorOff = 0, exteriorOff = 0;
+        for (auto iter = sdf->cbeginValueOff(); iter; ++iter) {
+            const float distance = float(ellipse.project(iter.getCoord().asVec3d()) * float(sdf->voxelSize()[0]));
+            const bool interior = (distance <= 0.0);
+            if (interior) EXPECT_EQ(-sdf->background(), *iter);
+            else          EXPECT_EQ(sdf->background(), *iter);
+            interior ? ++interiorOff : ++exteriorOff;
+        }
+
+        EXPECT_EQ(size_t(0), interiorOff);
+        EXPECT_EQ(size_t(85363), exteriorOff);
+    }
+}
+
 
 TEST_F(TestPointRasterizeSDF, testAttrTransfer)
 {
     // Test no points
     {
         float radius = 0.2f, search = 0.4f;
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({}).voxelsize(0.1).get();
         s.transform = nullptr;
@@ -920,7 +1339,7 @@ TEST_F(TestPointRasterizeSDF, testAttrTransfer)
 
     // Test 8 point attribute transfers for normal spheres and smooth spheres
     {
-        FixedSpheres<points::NullFilter> s;
+        FixedSurfacing<points::NullFilter> s;
 
         std::vector<Vec3f> positions = getBoxPoints(/*scale*/0.5f);
         const std::vector<int64_t> test1data = {9,10,11,12,13,14,15,16};
@@ -1016,7 +1435,7 @@ TEST_F(TestPointRasterizeSDF, testVariableAttrTransfer)
     // Test no points
     {
         float radius = 0.2f, search = 0.4f;
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
         s.halfband = 3;
         s.points = PointBuilder({}).voxelsize(0.1).get();
         s.transform = nullptr;
@@ -1043,7 +1462,7 @@ TEST_F(TestPointRasterizeSDF, testVariableAttrTransfer)
     // Test 8 point attribute transfers - overlapping spheres with varying
     // radii and a different target transform
     {
-        VariableSpheres<points::NullFilter> s;
+        VariableSurfacing<points::NullFilter> s;
 
         std::vector<Vec3f> positions = getBoxPoints(/*scale*/0.35f);
         const std::vector<int64_t> test1data = {0,1,2,3,4,5,6,7};
