@@ -12,6 +12,11 @@
 #include <openvdb/tree/LeafManager.h>
 #include <openvdb/util/logging.h>
 #include <openvdb/util/Assert.h>
+
+#if defined(_WIN32)
+#include <GL/glu.h>
+#endif
+
 #include <algorithm> // for std::min()
 #include <cmath> // for std::abs(), std::fabs(), std::floor()
 #include <limits>
