@@ -626,7 +626,11 @@ public:
 
     static OP_Node* factory(OP_Network*, const char* name, OP_Operator*);
 
-    int isRefInput(unsigned i ) const override { return (i > 0); }
+    int isRefInput(OP_InputIdx i) const override
+    {
+        UT_ASSERT(i >= 0);
+        return (i > 0);
+    }
 
     class Cache: public SOP_VDBCacheOptions
     {
